""" bitfield.py - bitfields using Python standard library. """

from math import ceil
from collections import namedtuple


class BitField():
    """BitField class -- Implements bitfields using the standard library.

    Attributes:
        size (int) - size of the bit field
        bitfield (bytearray) - byte array containing the bitfield
        Position (namedtuple) - Named tuple representing a bit's location
                                within the bytearray.
    """
    def __init__(self, size):
        self.size = size  # TODO bounds checking
        self.bitfield = bytearray([0x00] * int(ceil(size / 8)))
        self.position = namedtuple("position", ["byte", "bit"])

    def setbit(self, position):
        """BitField.setbit() - set bit at specified position to 1

        Args:
            position (int) - Position to set.

        Returns:
            Nothing.
        """
        pos = self.getpos(position)
        self.bitfield[pos.byte] |= (0x01 << pos.bit) & 0xff

    def unsetbit(self, position):
        """BitField.unsetbit() - set bit at specified position to 0

        Args:
            position (int) - Position to unset.

        Returns:
            Nothing.
        """
        pos = self.getpos(position)
        self.bitfield[pos.byte] &= ~(0x01 << pos.bit) & 0xff

    def getbit(self, position):
        """Bitfield.getbit() - Retrieve contents of bit at a specific location.

        Args:
            position (int) - Position to retrieve.

        Returns:
            True if bit is set (1).
            False if bit is not set (0).
        """
        pos = self.getpos(position)
        if self.bitfield[pos.byte] & ((0x01 << pos.bit) & 0xff):
            return True
        return False

    def zero(self):
        """Bitfield.zero() - Set all bits to zero.

        Args:
            None

        Returns:
            Nothing
        """
        for position in range(len(self.bitfield)):
            self.bitfield[position] = 0x00

    def one(self):
        """Bitfield.one() - Set all bits to one.

        Args:
            None

        Returns:
            Nothing
        """
        for position in range(len(self.bitfield)):
            self.bitfield[position] = 0xff

    def getpos(self, position):
        """Bitfield.getpos() - Get position of a bit in a bitfield.

        Args:
            position (int) - Position to retrieve.

        Returns:
            position (namedtuple) containing byte and bit positions.

        Example:
            I want to get the position of the 100th bit. Since this is stored
            in a bytearray, one can't just do something like this:
                value = bitfield[100] # This will get the 100th byte, not bit!

            The 100th bit of a bytearray will be 4 bits into the 12th byte:
                >>> bitfield.getpos(100)
                Position(byte=12, bit=4)
        """
        bytepos = int(ceil(position / 8)) - 1
        bitpos = position % 8
        if bitpos != 0:
            bitpos = 8 - bitpos
        return self.position(bytepos, bitpos)



""" bloomfilter.py - Bloom filters using Python standard library. """

from math import ceil, log

try:
    import mmh3
except ImportError:
    import dmfrbloom.pymmh3 as mmh3

from .bitfield import BitField


class BloomFilter():
    """BloomFilter class - Implements bloom filters using the standard library.

        Attributes:
            size (int) - size of the filter in bits.
            hashcount (int) - number of hashes per element.
            filter - (BitField object) - bitfield containing the filter.
    """
    def __init__(self, expected_items, fp_rate):
        self.size = self.ideal_size(expected_items, fp_rate)
        self.hashcount = self.ideal_hashcount(expected_items)
        self.filter = BitField(self.size)

    def add(self, element):
        """BloomFilter.add() - Add an element to the filter.

        Args:
            element (str) - Element to add to the filter.

        Returns:
            Nothing.
        """
        for seed in range(self.hashcount):
            result = mmh3.hash(str(element), seed) % self.size
            self.filter.setbit(result)

    def lookup(self, element):
        """BloomFilter.lookup() - Check if element exists in the filter.

        Args:
            element (str) - Element to look up.

        Returns:
            Nothing.
        """
        for seed in range(self.hashcount):
            result = mmh3.hash(str(element), seed) % self.size
            if self.filter.getbit(result) is False:
                return False
        return True

    def save(self, path):
        """BloomFilter.save() - Save the filter's current state to a file.

        Args:
            path (str) - Location to save the file

        Returns:
            Nothing.

        TODO: error checking if file cant be written.
        """
        with open(path, "wb") as filterfile:
            filterfile.write(self.size.to_bytes(16, byteorder="little"))
            filterfile.write(self.hashcount.to_bytes(16, byteorder="little"))
            filterfile.write(self.filter.bitfield)

    def load(self, path):
        """BloomFilter.load() - Load a saved filter.

        Args:
            path (str) - Location of filter to load.

        TODO: error check if this exists + is readable!
        """
        with open(path, "rb") as filterfile:
            self.size = int.from_bytes(filterfile.read(16), byteorder="little")
            self.hashcount = \
                int.from_bytes(filterfile.read(16), byteorder="little")
            self.filter.bitfield = filterfile.read()

    @staticmethod
    def accuracy(size, hashcount, elements):
        """BloomFilter.accuracy() - Calculate a filter's accuracy given
                                    size, hash count, and expected number
                                    of elements.

        Args:
            size (int) - Size of filter.
            hashcount (int) - Number of hashes per element.
            elements (int) - Number of expected elements.

        Returns:
            float containing a filter's percentage of accuracy.
        """
        # fp = (1 - [1 - 1 / size] ^ hashcount * expected_items ^ hashcount
        false_positive = \
            (1 - (1 - 1 / size) ** (hashcount * int(elements))) \
            ** hashcount
        print('FALSE: ', false_positive)
        return round(100 - false_positive * 100, 4)

    @staticmethod
    def ideal_size(expected, fp_rate):
        """BloomFilter.ideal_size() - Calculate ideal filter size given an
                                      expected number of elements and desired
                                      rate of false positives.

        Args:
            expected (int) - Expected number of elements in the filter.
            fp_rate (int) - Acceptable rate of false positives. Ex: 0.01 will
                            tolerate 0.01% chance of false positives.

        Returns:
            Ideal size.
        """
        return int(-(expected * log(fp_rate)) / (log(2) ** 2))

    def ideal_hashcount(self, expected):
        """BloomFilter.ideal_hashcount() - Calculate ideal number of hashes to
                                           perform given the expected number of
                                           elements to be stored in the filter.

        Args:
            expected (int) - Expected number of elements.

        Returns:
            Ideal number of hashes (int).
        """
        # ideal = (size / expected items) * log(2)
        return int((self.size / int(expected)) * log(2))

    @property
    def bytesize(self):
        """BloomFilter.bytesize() - Get size of filter.

        Args:
            None

        Returns:
            Size of filter (int)
        """
        return ceil(self.size / 8)

    @property
    def bytesize_human(self):
        """BloomFilter.bytesize_human() - Get human-readable size of filter.

        Stole this from someone on StackOverflow but don't remember where.

        Args:
            None.

        Returns:
            Filter size (str).
        """
        suffix = ['bytes', 'Kb', 'Mb', 'Gb', 'Tb', 'Pb', 'Eb', 'Zb', 'Yb']
        order = int(log(ceil(self.size) / 8, 2) / 10) if self.size else 0
        rounded = round(ceil(self.size / 8) / (1 << (order * 10)), 4)
        return str(rounded) + suffix[order]



""" timefilter.py - Time-decay bloom filters using Python standard library."""

import time
from math import log

try:
    import mmh3
except ImportError:
    import dmfrbloom.pymmh3 as mmh3


class TimeFilter():
    """TimeFilter class - Implements time-decaying bloom filters.

    Attributes:
        size (int)      - Size of the filter.
        hashcount (int) - Ideal number of hashes per filter element.
        filter (list)   - List of elements. Value is None or a timestamp.
        timeout (int)   - Seconds that elements should be valid.

    Args:
        expected (int)   - Expected number of elements.
        f_rate (float)   - Acceptable false positve rate. Ex: 0.001
        timeout (int)    - Number of second an element is valid for.
        precision (bool) - Use floating point for timestamps? True/False
    """
    def __init__(self, expected, fp_rate, timeout, precision=False):
        self.size = self.ideal_size(expected, fp_rate)
        self.hashcount = self.ideal_hashcount(self.size, expected)
        self.filter = [None] * self.size
        self.timeout = timeout
        self.precision = precision

    @staticmethod
    def ideal_size(expected, fp_rate):
        """ideal_size() - Calculate ideal filter size.
        Args:
            expected (int) - Expected number of elements to add to the filter.
            fp_rate (int) - Desired false positive rate. Ex: 0.01 for 99.99%
        Returns:
            Ideal size (int)
        """
        return int(-(expected * log(fp_rate)) / (log(2) ** 2))

    @staticmethod
    def ideal_hashcount(size, expected):
        """ideal_hashcount() - Calculate ideal number of hashes per element.
        Args:
            size (int) - Size of the filter.
            expected (int) - Expected number of elements.
        Returns:
            Ideal number of hashes to use per element.
        """
        return int((size / int(expected)) * log(2))

    def add(self, element):
        """add() - Add an element to the filter.
        Args:
            element (str) - Element to add to the filter.
        Returns:
            Nothing.
        """
        current_time = time.time() if self.precision else int(time.time())
        for seed in range(self.hashcount):
            result = mmh3.hash(str(element), seed) % self.size
            self.filter[result] = current_time

    def lookup(self, element):
        """lookup() - Check if an element exists in the filter. This also
                      reaps expired elements if they are found.
        Args:
            element (str) - Element to check for.
        Returns:
            True if element is likely to exist in the filter.
            False if the element definitely does not exist in the filter.
        """
        current_time = time.time()
        result = True
        for seed in range(self.hashcount):
            index = mmh3.hash(str(element), seed) % self.size
            if self.filter[index] is None:
                result = False
            elif current_time - self.filter[index] > self.timeout:
                self.filter[result] = None
                result = False
        return result
